@@ -40,7 +40,7 @@ from .utils.notebook import scale_svg
 from .utils.project import expand_path
 
 
-class WindowDict(collections.MutableMapping):
+class WindowDict(collections.abc.MutableMapping):
     '''
     Dictionary for holding window objects
 
@@ -63,7 +63,7 @@ class WindowDict(collections.MutableMapping):
     '''
 
     def __init__(self, *args, **kwargs):
-        collections.MutableMapping.__init__(self, *args, **kwargs)
+        collections.abc.MutableMapping.__init__(self, *args, **kwargs)
         self._data = dict()
         self.project = None
         self.project_handle = None
@@ -155,7 +155,10 @@ class WindowDict(collections.MutableMapping):
     def __delitem__(self, key):
         del self._data[key]
         for window in self._data.values():
-            window.delete_target(key)
+            try:
+                window.delete_target(key)
+            except ValueError:
+                pass
 
     def __iter__(self):
         return iter(self._data)
@@ -209,7 +212,7 @@ class TemplateDict(WindowDict):
         del self._data[key]
 
 
-class ContinuousQuery(ESPObject, collections.MutableMapping):
+class ContinuousQuery(ESPObject, collections.abc.MutableMapping):
     '''
     Continuous Query
 
@@ -506,7 +509,7 @@ class ContinuousQuery(ESPObject, collections.MutableMapping):
                 for source in re.split(r'\s+', item.attrib.get('source', '').strip()):
                     if not source:
                         continue
-                    out.windows[source].add_target(target, role=item.get('role'),
+                    out.windows[source].add_target(out.windows[target], role=item.get('role'),
                                                    slot=item.get('slot'))
 
         for item in data.findall('./metadata/meta'):
