@@ -1,11 +1,11 @@
 import logging
+import struct
 import uuid
+import io
 import sys
 import matplotlib
 
-from base64 import b16encode
-
-import plotly.colors as clrs
+from base64 import b16encode, b64encode
 
 from matplotlib import cm
 
@@ -72,6 +72,17 @@ class Options(object):
     def items(self):
         return(self._opts.items())
 
+    def __str__(self):
+        i = 0
+        s = ""
+        for name,value in self._opts.items():
+            if i > 0:
+                s += ","
+            s += name
+            s += "="
+            s += value
+        return(s)
+
     @property
     def options(self):
         return(self._opts)
@@ -198,11 +209,15 @@ class Colors(Options):
             self.createFromColors(self.getOpt("colors"))
 
     def createFromColorMap(self,colormap):
+
+        if "clrs" not in sys.modules:
+            import plotly.colors as clrs
+
         colors = []
         colorscale = []
         luma = []
 
-        if colormap != None:
+        if colormap != None and len(colormap) > 0:
             if colormap.find("sas_") == 0:
                 if colormap in Colors._sasThemes:
                     colors.extend(Colors._sasThemes[colormap])
@@ -514,10 +529,27 @@ def indexOf(list,o):
 
 def addTo(list,o):
     if contains(list,o) == False:
-        list.append(o);
-        return(True);
+        list.append(o)
+        return(True)
 
-    return(False);
+    return(False)
+
+def setItem(list,o):
+    item = None
+
+    for i in list:
+        if i == o:
+            item = i
+            break
+
+    if item == None:
+        item = o
+        list.append(o)
+    else:
+        for x in o:
+            item[x] = o[x]
+
+    return(item)
 
 def removeFrom(list,o):
     code = False
