@@ -9,6 +9,8 @@ import esppy.espapi.viewers as viewers
 
 from esppy.espapi.tools import Options, Colors
 
+from base64 import b64encode
+
 import sys
 
 import threading
@@ -48,7 +50,7 @@ class Visuals(Options):
         elif self.hasOpt("colors"):
             self._colors = tools.Colors(colors=self.getOpt("colors"))
         else:
-            self._colors = tools.Colors(colormap=None)
+            self._colors = tools.Colors(colormap="")
 
         self._titleStyle = tools.Options(font_size="12pt")
 
@@ -73,6 +75,13 @@ class Visuals(Options):
         self._visuals.append(chart)
         return(chart)
 
+    def createScatterPlot(self,datasource,**kwargs):
+        datasource.addDelegate(self)
+        chart = ScatterPlot(self,datasource,**kwargs)
+        chart.create()
+        self._visuals.append(chart)
+        return(chart)
+
     def createTimeSeries(self,datasource,**kwargs):
         datasource.addDelegate(self)
         chart = TimeSeries(self,datasource,**kwargs)
@@ -122,6 +131,13 @@ class Visuals(Options):
         self._visuals.append(chart)
         return(chart)
 
+    def createImageViewer(self,datasource,**kwargs):
+        datasource.addDelegate(self)
+        chart = ImageViewer(self,datasource,**kwargs)
+        chart.create()
+        self._visuals.append(chart)
+        return(chart)
+
     def createImages(self,datasource,**kwargs):
         datasource.addDelegate(self)
         chart = Images(self,datasource,**kwargs)
@@ -441,15 +457,22 @@ class LineChart(Chart):
 
         colors = self._visuals._colors.getFirst(len(values))
 
+        mode = "lines"
+
+        if fill:
+            mode = "none"
+        elif self.hasOpt("mode"):
+            mode = self.getOpt("mode")
+
         for i,v in enumerate(values):
             if fill:
                 if i == 0:
-                    self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="none",fill="tozeroy",fillcolor=colors[i]))
+                    self._data.append(go.Scatter(x=[""],y=[0],name=v,mode=mode,fill="tozeroy",fillcolor=colors[i]))
                 else:
-                    self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="none",fill="tonexty",fillcolor=colors[i]))
+                    self._data.append(go.Scatter(x=[""],y=[0],name=v,mode=mode,fill="tonexty",fillcolor=colors[i]))
             else:
                 line["color"] = colors[i]
-                self._data.append(go.Scatter(x=[""],y=[0],name=v,mode="lines",line=line))
+                self._data.append(go.Scatter(x=[""],y=[0],name=v,mode=mode,line=line))
 
     def draw(self,data = None,clear = False):
         if self._figure == None:
@@ -484,6 +507,11 @@ class LineChart(Chart):
 
         self.setTitle()
 
+class ScatterPlot(LineChart):
+    def __init__(self,visuals,datasource,**kwargs):
+        LineChart.__init__(self,visuals,datasource,**kwargs)
+        self.setOpt("mode","markers")
+
 class TimeSeries(LineChart):
     def __init__(self,visuals,datasource,**kwargs):
         LineChart.__init__(self,visuals,datasource,**kwargs)
@@ -769,6 +797,51 @@ class Table(Chart):
 
         self.setTitle()
 
+class ImageViewer(Chart):
+    def __init__(self,visuals,datasource,**kwargs):
+        Chart.__init__(self,visuals,datasource,**kwargs)
+        self._data = None
+        self._detection = None
+
+    def draw(self,data = None,clear = False):
+        if data != None and len(data) > 0:
+
+            if self._detection == None:
+                if self._datasource.schema.hasFields():
+                    self._detection = (self._datasource.schema.getField("_nObjects_") != None)
+
+            self._data = data[len(data) - 1]
+            field = self.getOpt("image");
+
+            html = None
+
+            if field in self._data:
+                imagedata = b64encode(self._data[field]).decode("utf-8")
+                html = ""
+                html += "<div style='width:" + str(self.getOpt("image_width",400)) + "px;height:" + str(self.getOpt("image_height",400)) + "px;position:relative;margin:auto;border:" + self.getOpt("image_border","1px solid #000000") + "'>"
+                html += "<img style='width:100%;height:100%' src='data:image/jpeg;base64," + imagedata + "'/>"
+
+                if self._detection:
+                    if "_nObjects_" in self._data:
+                        value = str(self._data["_nObjects_"])
+                        numObjects = int(float(value))
+                        for i in range(0,numObjects):
+                            s = "_Object" + str(i) + "_"
+                            text = self._data[s].strip()
+                            s = "_Object" + str(i) + "_x"
+                            x = int(float(self._data[s]) * 100)
+                            s = "_Object" + str(i) + "_y"
+                            y = int(float(self._data[s]) * 100)
+                            html += "<div style='position:absolute;zindex:1000;font-weight:normal;color:" + self.getOpt("label_color","white") + ";left:" + str(x) + "%;top:" + str(y) + "%;'>" + text + "</div>"
+
+                html += "</div>";
+
+            if html != None:
+                content = widgets.HTML(value=html,layout=widgets.Layout(border=self._visuals.getOpt("border","1px solid #d8d8d8"),width="100%",height="100%",overflow="auto"))
+                self.children = [self._banner,content]
+
+        self.setTitle()
+
 class Images(Chart):
 
     def __init__(self,visuals,datasource,**kwargs):
@@ -812,12 +885,12 @@ class Images(Chart):
         keys = []
 
         for o in data:
-            if ("_key_" in o) == False:
+            if ("@key" in o) == False:
                 continue
 
-            key = o["_key_"]
+            key = o["@key"]
 
-            if "_opcode" in o and o["_opcode"] == "delete":
+            if "@opcode" in o and o["@opcode"] == "delete":
                 if self.removeEntry(key) != None:
                     layout = True
             else:
@@ -907,15 +980,18 @@ class ImageEntry(Options):
             if field in self._data:
                 data = self._data[field]
 
-                if data.find(Visuals._dataHeader) == 0:
+                #if data.find(Visuals._dataHeader) == 0:
+                if True:
                     s = data[len(Visuals._dataHeader):]
-                    key = self._data["_key_"]
+                    key = self._data["@key"]
                     index = s.find(":")
                     format = s[0:index]
                     s = s[index + 1:]
 
                     html += "<div style='width:" + str(self._images.getOpt("image_width",400)) + "px;height:" + str(self._images.getOpt("image_height",400)) + "px;position:relative;margin:auto;border:" + self._images.getOpt("image_border","1px solid #000000") + "'>"
-                    html += "<img style='width:100%;height:100%' src='data:image/" + format + ";base64," + s + "'/>"
+                    #html += "<img style='width:100%;height:100%' src='data:image/" + format + ";base64," + s + "'/>"
+                    #html += "<img style='width:100%;height:100%' src='data:application/octet-stream;base64," + s + "'/>"
+                    html += "<img style='width:100%;height:100%' src='data:image/jpeg;base64," + data + "'/>"
 
                     if self._images._detection:
                     #if False:
@@ -950,8 +1026,8 @@ class ImageEntry(Options):
     @data.setter
     def data(self,value):
         self._data = value
-        if "_key_" in self._data:
-            self._key = self._data["_key_"]
+        if "@key" in self._data:
+            self._key = self._data["@key"]
 
 class Map(Chart):
     def __init__(self,visuals,datasource,**kwargs):
@@ -1112,7 +1188,7 @@ class Map(Chart):
         keyValues = self.getValues("keys")
 
         if len(keyValues) == 0:
-            keyValues = ["_key_"]
+            keyValues = ["@key"]
 
         keys = []
 
@@ -1176,7 +1252,7 @@ class Map(Chart):
                 for i,s in enumerate(popup):
                     if i > 0:
                         text += "<br/>"
-                    text += s + "=" + value[s]
+                    text += s + "=" + str(value[s])
                 marker.popup.value = text
 
         remove = {}
@@ -1434,11 +1510,11 @@ class Gauge(Chart):
             layout = True
 
         for o in data:
-            if ("_key_" in o) == False:
+            if ("@key" in o) == False:
                 continue
-            key = o["_key_"]
+            key = o["@key"]
 
-            if "_opcode" in o and o["_opcode"] == "delete":
+            if "@opcode" in o and o["@opcode"] == "delete":
                 if key in self._gauges:
                     self._gauges.pop(key)
                     layout = True
@@ -1713,12 +1789,12 @@ class Compass(Chart):
             layout = True
 
         for o in data:
-            if ("_key_" in o) == False:
+            if ("@key" in o) == False:
                 continue
 
-            key = o["_key_"]
+            key = o["@key"]
 
-            if "_opcode" in o and o["_opcode"] == "delete":
+            if "@opcode" in o and o["@opcode"] == "delete":
                 if key in self._entries:
                     self._entries.pop(key)
                     layout = True
