@@ -36,7 +36,7 @@ from ..utils.data import gen_name
 from ..utils.notebook import scale_svg
 
 
-class WindowDict(collections.MutableMapping):
+class WindowDict(collections.abc.MutableMapping):
     '''
     Dictionary for holding window objects
 
@@ -61,7 +61,7 @@ class WindowDict(collections.MutableMapping):
     '''
 
     def __init__(self, template, *args, **kwargs):
-        collections.MutableMapping.__init__(self, *args, **kwargs)
+        collections.abc.MutableMapping.__init__(self, *args, **kwargs)
         self._data = dict()
         self.template = template
         self.project = None
@@ -189,7 +189,10 @@ class WindowDict(collections.MutableMapping):
     def __delitem__(self, key):
         del self._data[key]
         for window in self._data.values():
-            window.delete_target(key)
+            try:
+                window.delete_target(key)
+            except ValueError:
+                pass
 
     def __iter__(self):
         return iter(self._data)
@@ -204,7 +207,7 @@ class WindowDict(collections.MutableMapping):
         return repr(self._data)
 
 
-class Template(ESPObject, collections.MutableMapping):
+class Template(ESPObject, collections.abc.MutableMapping):
     '''
     ESP Template
 
@@ -511,7 +514,7 @@ class Template(ESPObject, collections.MutableMapping):
             raise TypeError('Only CalculationWindow and TrainWindow objects support the method')
         return window.set_parameters(**parameters)
 
-    def set_inputs(self, window=None, **input_map):
+    def set_inputs(self, window=None, model=None, **input_map):
         '''
         Set inputs
 
@@ -519,6 +522,8 @@ class Template(ESPObject, collections.MutableMapping):
         ----------
         window : Window, optional
             The Window object to set inputs, default value is None
+        model : string
+            The name / URL of the model
         **input_map : keyword-arguments, optional
             The parameters to set
 
@@ -538,11 +543,14 @@ class Template(ESPObject, collections.MutableMapping):
             raise ValueError('%s is not a one of Template %s' %
                              (base_name, self.name))
 
-        if not isinstance(window, (TrainWindow, CalculateWindow, ScoreWindow)):
+        if isinstance(window, (TrainWindow, CalculateWindow)):
+            return window.set_inputs(**input_map)
+        elif isinstance(window, ScoreWindow):
+            return window.set_inputs(model, **input_map)
+        else:
             raise TypeError('Only CalculationWindow, TrainWindow and ScoreWindow objects support the method')
-        return window.set_inputs(**input_map)
 
-    def set_outputs(self, window=None, **output_map):
+    def set_outputs(self, window=None, model=None, **output_map):
         '''
         Set outputs
 
@@ -550,6 +558,8 @@ class Template(ESPObject, collections.MutableMapping):
         ----------
         window : Window, optional
             The Window object to set outputs, default value is None
+        model : string
+            The name / URL of the model
         **output_map : keyword-arguments, optional
             The parameters to set
 
@@ -568,9 +578,12 @@ class Template(ESPObject, collections.MutableMapping):
             raise ValueError('%s is not a one of Template %s' %
                              (base_name, self.name))
 
-        if not isinstance(window, (CalculateWindow, ScoreWindow)):
+        if isinstance(window, (CalculateWindow, ScoreWindow)):
+            return window.set_outputs(**output_map)
+        elif isinstance(window, ScoreWindow):
+            return window.set_outputs(model, **output_map)
+        else:
             raise TypeError('Only CalculationWindow and ScoreWindow objects support the method')
-        return window.set_outputs(**output_map)
 
     def add_target(self, obj, **kwargs):
         '''
